module SubscribeToChannelTests exposing (..)

import Expect
import Test exposing (..)
import Elmer
import Elmer.Subscription as Subscription
import TestHelpers as Helpers exposing (Msg(..))
import Procedure
import Procedure.Channel as Channel


subscribeTests : Test
subscribeTests =
  describe "wait on a subscription to a channel" <|
  let
    procedureState =
      Helpers.tryProcedure (\_ ->
            Procedure.fetch (Helpers.stringCommand "First")
              |> Procedure.andThen (\result -> 
                Channel.join (Helpers.stringSubscription result)
                  |> Channel.acceptOne
              )
              |> Procedure.andThen (\result -> Procedure.fetch <| Helpers.stringCommand <| "After sub: " ++ result)
              |> Procedure.map (\result -> "Mapped: " ++ result)
          )
        |> Subscription.with (\_ -> Helpers.testSubscriptions)
        |> Subscription.send "string-subscription" "value from subscription"
  in
  [ test "it waits for the subscription and then continues the procedure" <|
    \() ->
      procedureState
        |> Helpers.expectResult (Ok "Mapped: After sub: First then value from subscription")
  , describe "when the subscription is received"
    [ test "it removes the subscription" <|
      \() ->
        procedureState
          |> Elmer.expectModel (\model ->
            Helpers.testSubscriptions model
              |> Expect.equal Sub.none
          )
    ]
  ]


subscribeAndFilterTests : Test
subscribeAndFilterTests =
  describe "when a subscription expects a particular value"
  [ describe "when the value is received"
    [ test "it processes the remainder of the procedure" <|
      \() ->
        Helpers.tryProcedure (\_ ->
          Procedure.provide "sub-key"
            |> Procedure.andThen (\result ->
              Channel.join Helpers.keySubscription
                |> Channel.filter (\_ desc -> desc.key == result)
                |> Channel.acceptOne
            )
            |> Procedure.map .value
            |> Procedure.andThen (\result -> Procedure.fetch <| Helpers.stringCommand <| "After sub: " ++ result)
            |> Procedure.map (\result -> "Mapped: " ++ result)
        )
          |> Subscription.with (\_ -> Helpers.testSubscriptions)
          |> Subscription.send "key-subscription" { key = "sub-key", value = "awesome value" }
          |> Helpers.expectResult (Ok "Mapped: After sub: awesome value")
    ]
  , describe "when the wrong value is received"
    [ test "it ignores that value and continues to wait" <|
      \() ->
        Helpers.tryProcedure (\_ ->
          Procedure.provide "sub-key"
            |> Procedure.andThen (\result ->
              Channel.join Helpers.keySubscription
                |> Channel.filter (\_ desc -> desc.key == result)
                |> Channel.acceptOne
            )
            |> Procedure.map .value
            |> Procedure.andThen (\result -> Procedure.fetch <| Helpers.stringCommand <| "After sub: " ++ result)
            |> Procedure.map (\result -> "Mapped: " ++ result)
        )
          |> Subscription.with (\_ -> Helpers.testSubscriptions)
          |> Subscription.send "key-subscription" { key = "wrong-key", value = "awesome value" }
          |> Subscription.with (\_ -> Helpers.testSubscriptions)
          |> Subscription.send "key-subscription" { key = "sub-key", value = "fun value" }
          |> Helpers.expectResult (Ok "Mapped: After sub: fun value")
    ]
  ]


filterOnProcedureIdTests : Test
filterOnProcedureIdTests =
  describe "when filtering for a specific value" <|
  let
    testState =
      Helpers.tryProcedure (\_ ->
        Procedure.provide "something"
          |> Procedure.andThen (\result ->
            Channel.join Helpers.keySubscription
              |> Channel.filter (\channelId desc -> desc.key == channelId)
              |> Channel.acceptOne
          )
          |> Procedure.map .value
      )
        |> Helpers.andTryProcedure (\_ ->
            Procedure.provide "something else"
              |> Procedure.andThen (\result ->
                  Channel.join Helpers.intSubscription
                    |> Channel.filter (\channelId number -> String.contains (String.fromInt number) channelId)
                    |> Channel.acceptOne
              )
              |> Procedure.map String.fromInt
          )
        |> Subscription.with (\_ -> Helpers.testSubscriptions)
  in
  [ test "it provides the first procedure with the procedureId" <|
    \() ->
      testState
        |> Subscription.send "key-subscription" { key = "0-0", value = "awesome value" }
        |> Helpers.expectResult (Ok "awesome value")
  , test "it provides the second procedure with the procedureId" <|
    \() ->
      testState
        |> Subscription.send "int-subscription" 1
        |> Helpers.expectResult (Ok "1")
  ]


multipleChannelTests : Test
multipleChannelTests =
  describe "when procedures with different channels are running" <|
  let
    testState =
      Helpers.tryProcedure (\_ ->
        Procedure.fetch (Helpers.stringCommand "First")
          |> Procedure.andThen (\result -> 
            Channel.join (Helpers.stringSubscription result)
              |> Channel.acceptOne
          )
          |> Procedure.andThen (\result -> Procedure.fetch <| Helpers.stringCommand <| "After sub: " ++ result)
          |> Procedure.map (\result -> "Mapped: " ++ result)
      )
        |> Helpers.andTryProcedure (\_ ->
        Procedure.fetch (Helpers.stringCommand "Second")
          |> Procedure.andThen (\result -> 
            Channel.join Helpers.intSubscription
              |> Channel.acceptOne
          )
          |> Procedure.andThen (\result ->
            Procedure.fetch <| Helpers.stringCommand <| "After sub: " ++ String.fromInt result
          )
          |> Procedure.map (\result -> "Mapped: " ++ result)
        )
  in
  [ describe "when the first subscription is received" <|
    let
      firstSubState =
        testState
          |> Subscription.with (\_ -> Helpers.testSubscriptions)
          |> Subscription.send "string-subscription" "value from subscription"
    in
    [ test "it completes the first procedure" <|
      \() ->
        firstSubState
          |> Helpers.expectResult (Ok "Mapped: After sub: First then value from subscription")
    , describe "when the second subscription is received" <|
      let
        secondSubState =
          firstSubState
            |> Subscription.with (\_ -> Helpers.testSubscriptions)
            |> Subscription.send "int-subscription" 38
      in
      [ test "it completes the second procedure" <|
        \() ->
          secondSubState
            |> Helpers.expectResult (Ok "Mapped: After sub: 38")
      ]
    ]
  ]


mapTwoChannelsTests : Test
mapTwoChannelsTests =
  describe "when the procedure is waiting on two channels"
  [ test "it processes the channels sequentially" <|
    \() ->
      Helpers.tryProcedure (\_ ->
        Procedure.map2 (\a b -> a ++ " AND " ++ b)
          ( Procedure.fetch (Helpers.stringCommand "First String Command")
              |> Procedure.andThen (\result ->
                Channel.join (Helpers.stringSubscription result)
                  |> Channel.acceptOne
              )
              |> Procedure.andThen (\result -> Procedure.fetch <| Helpers.stringCommand <| "After sub: " ++ result)
          )
          ( Procedure.provide 787
              |> Procedure.andThen (\_ -> 
                Channel.join Helpers.intSubscription
                  |> Channel.acceptOne
              )
              |> Procedure.map (\result -> "Mapped Int: " ++ String.fromInt result)
          )
      )
        |> Subscription.with (\_ -> Helpers.testSubscriptions)
        |> Subscription.send "string-subscription" "value from subscription"
        |> Subscription.with (\_ -> Helpers.testSubscriptions)
        |> Subscription.send "int-subscription" 38
        |> Helpers.expectResult (Ok "After sub: First String Command then value from subscription AND Mapped Int: 38")
  ]
